# coding=utf-8
import datetime
import logging

from ..bigquery_mixin import BigQueryMixin
from .base_db_index import BaseMLIndex
from .cache import DatastoreTreeCache, GAEDatastoreTreeCache
from ..db import DatastoreConnection, GAEDatastoreConnection


# noinspection SqlNoDataSourceInspection,SqlResolve
class BigQueryMLIndex(BaseMLIndex, BigQueryMixin):
    STAGING_INDEX_TABLE_NAME = 'staging_index'
    INDEX_TABLE_NAME = 'index'

    def __init__(self, connection, version=None):
        self.__version = version or 0
        self.__create_tree_cache(connection)
        super(BigQueryMLIndex, self).__init__(connection)

    def __create_tree_cache(self, connection):
        if GAEDatastoreTreeCache is not None:
            datastore_connection = GAEDatastoreConnection(connection.data_volume_config)
        else:
            datastore_connection = DatastoreConnection(connection.data_volume_config, connection.project)

        tree_cache_class = GAEDatastoreTreeCache if GAEDatastoreTreeCache is not None else DatastoreTreeCache
        self.__tree_cache = tree_cache_class(datastore_connection)

    def __get_index_table_name(self, name):
        table_full_name = '{prefix}_{name}'.format(prefix=self._connection.table_prefix, name=name)

        return table_full_name

    def __get_index_table(self):
        return self._get_table(self.INDEX_TABLE_NAME, schema=self.__table_schema())

    def __get_staging_index_table(self):
        staging_table_name = '%s_%s' % (self.STAGING_INDEX_TABLE_NAME, self.__version)
        return self._get_table(staging_table_name, schema=self.__table_schema())

    def staging_count_items(self, with_metadata=False):
        staging_table = self.__get_staging_index_table()
        staging_index_table_full_name = staging_table.name

        with self._connection.get_cursor() as bq_dataset:
            query = """
                SELECT EXACT_COUNT_DISTINCT(name)
                FROM [{dataset_name}.{index_table_name}]
                """.format(
                    dataset_name=bq_dataset.name,
                    index_table_name=staging_index_table_full_name)

            if not with_metadata:
                query += " WHERE RIGHT(name, 9) != '.metadata'"

            results = list(self._query_sync(query, use_legacy_sql=True))

            return results[0][0]

    @classmethod
    def __table_schema(cls):
        from google.cloud import bigquery

        schema = (
            bigquery.SchemaField('name', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('sha', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('ctime', 'FLOAT', 'REQUIRED'),
            bigquery.SchemaField('mtime', 'FLOAT', 'REQUIRED'),
            bigquery.SchemaField('mode', 'INTEGER', 'REQUIRED'),
            bigquery.SchemaField('uid', 'INTEGER', 'REQUIRED'),
            bigquery.SchemaField('gid', 'INTEGER', 'REQUIRED'),
            bigquery.SchemaField('size', 'INTEGER', 'REQUIRED'),
            bigquery.SchemaField('ts', 'TIMESTAMP', 'REQUIRED'),
        )

        return schema

    def _create_table_if_needed(self):
        import google.api.core.exceptions

        with self._connection.get_cursor() as bq_dataset:
            def create_table(method):
                index_table = method()

                try:
                    index_table.create()
                except google.api.core.exceptions.Conflict:
                    pass

                return index_table

            create_table(self.__get_index_table)
            create_table(self.__get_staging_index_table)

    def set_entries(self, entries):
        if not entries:
            return

        now = datetime.datetime.utcnow()

        rows = [row + (now, ) for row in self._decode_entries(entries)]

        logging.debug('inserting %s rows into bq', len(rows))

        table = self.__get_staging_index_table()
        self.__tree_cache.set_entries(entries)
        self.handle_bq_errors(table.insert_data(rows))

        logging.debug('inserted %s rows into bq', len(rows))

    def get_commit_id(self):
        return self.__tree_cache.get_commit_id()

    def __truncate_staging(self):
        logging.info('truncate index staging')

        staging_table = self.__get_staging_index_table()
        staging_table.delete()

    def begin_commit(self, tree_id):
        bq_client = self._connection

        staging_table = self.__get_staging_index_table()

        with bq_client.get_cursor() as bq_dataset:
            src_query = """
                #standardSQL
                SELECT
                  * EXCEPT(row_number)
                FROM (
                  SELECT
                    *,
                    ROW_NUMBER() OVER (PARTITION BY name ORDER BY ts DESC) row_number
                  FROM
                    `{dataset_name}.{staging_table_name}` )
                WHERE
                  row_number = 1
              """.format(
                dataset_name=bq_dataset.name,
                staging_table_name=staging_table.name,
            )

            src_query_parameters = ()

            self._copy_table_data(
                src_query, src_query_parameters, self.__get_index_table())

    def end_commit(self):
        self.__truncate_staging()



# coding=utf-8
from ...gae_datastore_mixin import GAEDatastoreMixin
from .tree_cache import TreeCache
from google.appengine.ext import ndb


class TreeCacheModel(ndb.Model):
    files = ndb.TextProperty(required=True)
    sha = ndb.StringProperty(required=True)

    @classmethod
    def _get_kind(cls):
        return 'TreeCache'


class GAEDatastoreTreeCache(GAEDatastoreMixin, TreeCache):
    def __init__(self, connection):
        super(GAEDatastoreTreeCache, self).__init__(connection)

    @property
    def _entity_kind(self):
        return 'TreeCache'

    def _set_files(self, trees_objects):
        inserts = []
        for path, tree in trees_objects.items():
            blob = '\n'.join(self.build_blob(tree))

            name = path.decode('utf8')

            entity = TreeCacheModel(key=self._build_key(name), files=blob.encode('utf8'), sha=tree.id.decode('utf8'))

            inserts.append(entity)

        ndb.put_multi(inserts)

    def get_commit_id(self):
        entity = self._build_key('').get()

        if entity is None:
            return None

        return entity.sha

    def _get_files(self, scans):
        def name_from_key(entity):
            entity_id = entity.key.id()
            return '' if entity_id == 1 else entity_id

        keys = [self._build_key(path) for path, _, _ in scans]

        cache_entities = ndb.get_multi(keys)

        return {name_from_key(entity): self._files_sha_array_to_dict(entity.files) for entity in cache_entities if entity is not None}



# -*- coding: utf-8 -*-
"""This module provides a pretty printer for lucene query tree.
"""
from .tree import BaseOperation, BaseGroup, SearchField


class _StickMarker:
    """Use in list between two elements that must stick together
    """

    def __len__(self):
        return 0


# a marker to avoid a new line between two elements
_STICK_MARKER = _StickMarker()


class Prettifier(object):
    """Class to generate a pretty printer.
    """

    def __init__(self, indent=4, max_len=80, inline_ops=False):
        """
        The pretty printer factory.

        :param int indent: number of space for indentation
        :param int max_len: maximum line length in number of characters.
            Prettyfier will do its best to keep inside those margin,
            but as it can only split on operators, it may not be possible.
        :param bool inline_ops: if False (default) operators are printed on a new line
          if True, operators are printed at the end of the line.
        """
        self.indent = indent
        self.prefix = " " * self.indent
        self.max_len = max_len
        self.inline_ops = inline_ops

    def _get_chains(self, element, parent=None):
        """return a list of string and list, and recursively

        An inner list represent a level of indentation
        A string is information from the level
        """
        if isinstance(element, BaseOperation):
            if not isinstance(parent, BaseOperation) or element.op == parent.op:
                # same level, this is just associativity
                num_children = len(element.children)
                for n, child in enumerate(element.children):
                    yield from self._get_chains(child, element)
                    if n < num_children - 1:
                        if self.inline_ops:
                            yield _STICK_MARKER
                        if element.op:
                            yield element.op
            else:
                # another operation, raise level
                new_level = []
                num_children = len(element.children)
                for n, child in enumerate(element.children):
                    new_level.extend(self._get_chains(child, element))
                    if n < num_children - 1:
                        if self.inline_ops:
                            new_level.append(_STICK_MARKER)
                        if element.op:
                            new_level.append(element.op)
                yield new_level
        elif isinstance(element, BaseGroup):
            # raise level
            yield "("
            yield list(self._get_chains(element.expr, element))
            if self.inline_ops:
                yield _STICK_MARKER
            yield ")"
        elif isinstance(element, SearchField):
            # use recursion on sub expression
            yield element.name + ":"
            yield _STICK_MARKER
            yield from self._get_chains(element.expr, element)
        else:
            # simple element
            yield str(element)

    def _count_chars(self, element):
        """Replace each element by the element and a count of chars in it (and recursively)

        This will help, compute if elements can stand on a line or not
        """
        if isinstance(element, list):
            with_counts = [self._count_chars(c)for c in element]
            # when counting we add a space for joining
            return with_counts, sum(n + 1 for c, n in with_counts) - 1
        else:
            return element, len(element)

    def _apply_stick(self, elements):
        last = None
        sticking = False
        for current in elements:
            if current == _STICK_MARKER:
                assert last is not None, "_STICK_MARKER should never be first !"
                sticking = True
            elif sticking:
                last += " " + current
                sticking = False
            else:
                if last is not None:
                    yield last
                last = current
        yield last

    def _concatenates(self, chain_with_counts, char_counts, level=0, in_one_liner=False):
        """taking the result of _get_chains after passing through _count_chars,
        arrange things, using newlines and indentation when necessary

        :return string: prettified expression
        """
        # evaluate if it's feasible in one-line
        one_liner = in_one_liner or char_counts < self.max_len - (self.indent * level)
        new_level = level if one_liner else level + 1
        elements = [
            self._concatenates(c, n, level=new_level, in_one_liner=one_liner)
            if isinstance(c, list)
            else c
            for c, n in chain_with_counts]
        elements = self._apply_stick(elements)
        prefix = self.prefix if level and not in_one_liner else ""
        join_char = " " if one_liner else ("\n" + prefix)
        return prefix + join_char.join(l for c in elements for l in c.split("\n"))

    def __call__(self, tree):
        """Pretty print the query represented by tree

        :param tree: a query tree using elements from :py:mod:`luqum.tree`
        """
        chains = list(self._get_chains(tree))
        chain_with_counts, total = self._count_chars(chains)
        return self._concatenates(chain_with_counts, total)


prettify = Prettifier()
"""prettify function with default parameters
"""



# coding=utf-8
import logging
import random

from mali_commands.legit import BigQuerySqlHelper
from mali_commands.legit.scam import tree_to_sql_parts, parse_query
from tests.base import BaseTest

SqlHelper = BigQuerySqlHelper


class TestMetadata(BaseTest):
    def validate_sql(self, sql, params, expected_result, sql_helper_class=None):
        params = params or {}
        full_syntax = sql.format(**params)

        sql_helper_class = sql_helper_class or SqlHelper

        logging.debug('validate syntax %s with %s', full_syntax, sql_helper_class)

        actual_result = tree_to_sql_parts(parse_query(full_syntax), sql_helper_class())

        if expected_result != actual_result:
            self.assertEqual(expected_result, actual_result)

    def testSeed_only(self):
        seed = self.some_random_shit_number_int63()

        where_result = None

        self.validate_sql('seed:{seed}', dict(seed=seed), ({'seed': seed}, where_result))

    def testRange(self):
        value1 = self.some_random_shit_number_int63()
        value2 = self.some_random_shit_number_int63()
        field_name = 'field1_' + self.some_random_shit_alpha()

        where_result = '(({field_name}>={value1})AND({field_name}<={value2}))'.format(field_name=field_name, value1=value1, value2=value2)
        vars_result = None

        self.validate_sql('{field_name}:[{value1} TO {value2}]', dict(field_name=field_name, value1=value1, value2=value2), (vars_result, where_result))

    def testRange_inclusive(self):
        value1 = self.some_random_shit_number_int63()
        value2 = self.some_random_shit_number_int63()
        field_name = 'field1_' + self.some_random_shit_alpha()

        where_result = '(({field_name}>{value1})AND({field_name}<{value2}))'.format(field_name=field_name, value1=value1, value2=value2)
        vars_result = None

        self.validate_sql('{field_name}:{{{value1} TO {value2}}}', dict(field_name=field_name, value1=value1, value2=value2), (vars_result, where_result))

    def testSeed_twice(self):
        seed1 = self.some_random_shit_number_int63()
        seed2 = self.some_random_shit_number_int63()

        where_result = None

        self.validate_sql('seed:{seed1} AND seed:{seed2}', dict(seed1=seed1, seed2=seed2), ({'seed': seed2}, where_result))

    def testBuildSql_andOrOp(self):
        field_name1 = 'field1_' + self.some_random_shit_alpha()
        field_name2 = 'field2_' + self.some_random_shit_alpha()

        value1 = 'val1_' + self.some_random_shit_alpha()
        value2 = 'val2_' + self.some_random_shit_alpha()

        result_vars = None

        for op in ['AND', 'OR']:
            result_where = '({field_name1}="{value1}"){op}({field_name2}="{value2}")'.format(
                field_name1=field_name1, field_name2=field_name2, value1=value1, value2=value2, op=op)
            result_build_sql = (result_vars, result_where)

            self.validate_sql(
                '{field_name1}:{value1} {op} {field_name2}:{value2}',
                dict(field_name1=field_name1, field_name2=field_name2, value1=value1, value2=value2, op=op), result_build_sql)

    def testBuildSql_singleFieldSingleInt(self):
        value_int = self.some_random_shit_number_int63()
        field_name = 'field_' + self.some_random_shit_alpha()

        result_vars = None

        for sign in ['', '+']:
            result_where = '({field_name}={value})'.format(field_name=field_name, value=value_int)
            result_build_sql = (result_vars, result_where)

            self.validate_sql('{field_name}:{sign}{value}', dict(field_name=field_name, sign=sign, value=value_int), result_build_sql)
            self.validate_sql('{field_name}:({sign}{value})', dict(field_name=field_name, sign=sign, value=value_int), result_build_sql)

        for sign in ['-']:
            result_where = '({field_name}=-{value})'.format(field_name=field_name, value=value_int)
            result_build_sql = (result_vars, result_where)

            self.validate_sql('{field_name}:{sign}{value}', dict(field_name=field_name, sign=sign, value=value_int), result_build_sql)
            self.validate_sql('{field_name}:({sign}{value})', dict(field_name=field_name, sign=sign, value=value_int), result_build_sql)

        for op in ['=', '>', '<', '>=', '<=']:
            result_where = '({field_name}{op}{value})'.format(field_name=field_name, op=op, value=value_int)

            result_build_sql = (result_vars, result_where)

            self.validate_sql('{field_name}:{op}{value}', dict(field_name=field_name, op=op, value=value_int), result_build_sql)
            self.validate_sql('{field_name}:({op}{value})', dict(field_name=field_name, op=op, value=value_int), result_build_sql)
            self.validate_sql('{field_name}:({op}{value})', dict(field_name=field_name, op=op, value=value_int), result_build_sql)

    def testBuildSql_multiFieldInt(self):
        value_int1 = self.some_random_shit_number_int63()
        value_int2 = self.some_random_shit_number_int63()
        field_name1 = 'field1_' + self.some_random_shit_alpha()

        result_vars = None

        for sign in ['', '+']:
            result_where = '(({field_name}={value1})OR({field_name}={value2}))'.format(
                field_name=field_name1, value1=value_int1, value2=value_int2)

            result_build_sql = (result_vars, result_where)

            self.validate_sql(
                '{field_name}:({sign}{value1} {sign}{value2})',
                dict(field_name=field_name1, sign=sign, value1=value_int1, value2=value_int2), result_build_sql)

    def testBuildSql_singleFieldSingleFloat(self):
        value_float = self.some_random_shit_number_int63() / 911.0
        field_name = 'field_' + self.some_random_shit_alpha()

        result_vars = None

        for sign in ['', '+']:
            result_where = '({field_name}={value})'.format(field_name=field_name, value=value_float)
            result_build_sql = (result_vars, result_where)

            self.validate_sql('{field_name}:{sign}{value}', dict(field_name=field_name, sign=sign, value=value_float), result_build_sql)
            self.validate_sql('{field_name}:({sign}{value})', dict(field_name=field_name, sign=sign, value=value_float), result_build_sql)

        for sign in ['-']:
            result_where = '({field_name}=-{value})'.format(field_name=field_name, value=value_float)
            result_build_sql = (result_vars, result_where)

            self.validate_sql('{field_name}:{sign}{value}', dict(field_name=field_name, sign=sign, value=value_float), result_build_sql)
            self.validate_sql('{field_name}:({sign}{value})', dict(field_name=field_name, sign=sign, value=value_float), result_build_sql)

        for op in ['=', '>', '<', '>=', '<=']:
            result_where = '({field_name}{op}{value})'.format(field_name=field_name, op=op, value=value_float)

            result_build_sql = (result_vars, result_where)
            self.validate_sql('{field_name}:{op}{value}', dict(field_name=field_name, op=op, value=value_float), result_build_sql)
            self.validate_sql('{field_name}:({op}{value})', dict(field_name=field_name, op=op, value=value_float), result_build_sql)

    def testBuildSql_singleFieldSingleString(self):
        value_str = 'val_' + self.some_random_shit_alpha()
        field_name = 'field_' + self.some_random_shit_alpha()

        result_vars = None

        result_where = '({field_name}="{value}")'.format(field_name=field_name, value=value_str)

        result_build_sql = (result_vars, result_where)

        self.validate_sql('{field_name}:{value}', dict(field_name=field_name, value=value_str), result_build_sql)
        self.validate_sql('{field_name}:({value})', dict(field_name=field_name, value=value_str), result_build_sql)

        self.validate_sql('{field_name}: {value}', dict(field_name=field_name, value=value_str), result_build_sql)
        self.validate_sql('{field_name}: ({value})', dict(field_name=field_name, value=value_str), result_build_sql)

        self.validate_sql('{field_name}: "{value}"', dict(field_name=field_name, value=value_str), result_build_sql)
        self.validate_sql('{field_name}: ("{value}")', dict(field_name=field_name, value=value_str), result_build_sql)

        for op in ['=', '>', '<', '>=', '<=']:
            result_where = '({field_name}{op}"{value}")'.format(field_name=field_name, op=op, value=value_str)

            result_build_sql = (result_vars, result_where)
            self.validate_sql('{field_name}:{op}"{value}"', dict(field_name=field_name, op=op, value=value_str), result_build_sql)
            self.validate_sql('{field_name}:({op}"{value}")', dict(field_name=field_name, op=op, value=value_str), result_build_sql)

    def testBigQuerySample_only(self):
        sample = random.random()
        where_result = '(RAND($seed)>{sample:.4g})'.format(sample=1.0 - sample)
        vars_result = None

        self.validate_sql(
            'sample:{sample}',
            dict(sample=sample),
            (vars_result, where_result),
            sql_helper_class=BigQuerySqlHelper)

    def testBigQuerySample(self):
        sample = random.random()
        field_name = 'field1_' + self.some_random_shit_alpha()
        value = 'val1_' + self.some_random_shit_alpha()

        where_result = '({field_name}="{field_value}")AND(RAND($seed)>{sample:.4g})'.format(
            field_name=field_name, field_value=value, sample=(1.0 - sample))

        vars_result = None

        self.validate_sql(
            'sample:{sample} AND {field_name}:{field_value}',
            dict(sample=sample, field_name=field_name, field_value=value), (vars_result, where_result),
            BigQuerySqlHelper)

    def testBigQuerySample_implicit(self):
        sample = random.random()
        field_name = 'field1_' + self.some_random_shit_alpha()
        value = 'val1_' + self.some_random_shit_alpha()

        where_result = '({field_name}="{field_value}")AND(RAND($seed)>{sample:.4g})'.format(
            field_name=field_name, field_value=value, sample=(1.0 - sample))

        vars_result = None

        self.validate_sql(
            'sample:{sample} {field_name}:{field_value}',
            dict(sample=sample, field_name=field_name, field_value=value),
            (vars_result, where_result),
            BigQuerySqlHelper)

    def testBigQuerySample_twice(self):
        sample1 = random.uniform(0., 0.5)
        sample2 = random.uniform(0.5, 1.0)
        where_result = '(RAND($seed)>{sample1:.4g})AND(RAND($seed)>{sample2:.4g})'.format(
            sample1=1.0 - sample1, sample2=1.0 - sample2)
        vars_result = None

        self.validate_sql('sample:{sample1} AND sample:{sample2}', dict(sample1=sample1, sample2=sample2), (vars_result, where_result))



# coding=utf-8
import json
import os
import random
import click
import errno
import requests
from mali_commands.legit.dulwich import porcelain
from mali_commands.legit.dulwich.index import _fs_to_tree_path, index_entry_from_stat
from mali_commands.legit.dulwich.objects import Blob
from mali_commands.commons import add_to_data_if_not_none, output_result
from mali_commands.data_volume import create_data_volume, with_repo, validate_repo_data_path
import time
import shutil


@click.group('data')
def data_commands():
    pass


ignore_files = ['.DS_Store']


def get_total_files_in_path(paths):
    total_files = 0

    for path in paths:
        path = __expend_and_validate_path(path)

        if os.path.isdir(path):
            for root, dirs, files in os.walk(path, followlinks=True):
                for file_ in files:
                    if file_ in ignore_files:
                        continue

                    total_files += 1
        elif path not in ignore_files:
            total_files += 1

    return total_files


def __expend_and_validate_path(path, expand_vars=True, validate_path=True, abs_path=True):
    path = os.path.expanduser(path)
    if expand_vars:
        path = os.path.expandvars(path)

    if abs_path:
        path = os.path.abspath(path)

    if validate_path:
        if not os.path.exists(path):
            click.echo('not found %s' % path, err=True)
            exit(1)
            return

    return path


def enumerate_path(paths):
    for path in paths:
        path = __expend_and_validate_path(path)
        if os.path.isfile(path):
            yield path

        for root, dirs, files in os.walk(path, followlinks=True):
            for file_ in files:
                if file_ in ignore_files:
                    continue

                yield os.path.join(root, file_)


@data_commands.command('map')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--dataPath', required=True)
@click.pass_context
def _cmd_add_dataPath(ctx, volumeid, datapath):
    result = ctx.obj.handle_api(ctx.obj, requests.get, "data_volumes/{volume_id}".format(volume_id=volumeid))

    datapath = __expend_and_validate_path(datapath)

    create_data_volume(
        result['id'],
        datapath,
        not result['embedded'],
        result.get('display_name'),
        result.get('description'))

    display_name = result.get('display_name', 'No display name provided')
    click.echo('Initialized data volume %s (%s)' % (result['id'], display_name))


@data_commands.command('create')
@click.option('--displayName', required=True)
@click.option('--description', required=False)
@click.option('--org', required=True)
@click.option('--dataPath', required=True)
@click.option('--linked/--embedded', is_flag=True, required=True)
@click.pass_context
def _cmd_create_data_volume(ctx, displayname, description, org, datapath, linked):
    data = {}

    if org == 'me':
        org = None

    add_to_data_if_not_none(data, displayname, "display_name")
    add_to_data_if_not_none(data, org, "org")
    add_to_data_if_not_none(data, description, "description")
    add_to_data_if_not_none(data, not linked, "embedded")

    result = ctx.obj.handle_api(ctx.obj, requests.post, "data_volumes", data)

    data_volume_id = result['id']

    datapath = __expend_and_validate_path(datapath)

    create_data_volume(data_volume_id, datapath, linked, displayname, description)

    click.echo('Initialized data volume %s' % data_volume_id)


@data_commands.command('commit')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--message', '-m', required=False)
@click.pass_context
def commit_data_volume(ctx, volumeid, message):
    with with_repo(ctx, volumeid) as repo:
        result = repo.commit(message)

        if 'commit_id' not in result:
            click.echo('no changeset detected', err=True)
            exit(1)

        output_result(ctx, result)


@data_commands.command('log')
@click.option('--path', '-p', required=False)
@click.pass_context
def log_data_volume(ctx, path):
    with with_repo(ctx, path) as r:
        for entry in r.get_walker():
            commit = entry.commit
            click.echo("%s %s" % (commit.id.decode('utf8')[:7], commit.message.decode('utf8')))


@data_commands.command('add')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--files', '-f', multiple=True)
@click.option('--commit', is_flag=True, required=False)
@click.option('--message', '-m', required=False)
@click.pass_context
def add_to_data_volume(ctx, volumeid, files, commit, message):
    if commit and not message:
        raise click.UsageError('missing --message when using --commit')

    total_files = get_total_files_in_path(files)
    batch_size = min(total_files // 100, 2000)  # FIXME: hardcoded

    def add_batch(current_batch):
        porcelain.add(repo, current_batch, repo.data_volume_config.embedded)
        bar.update(len(current_batch))

    with with_repo(ctx, volumeid) as repo:
        with click.progressbar(length=total_files, show_pos=True) as bar:
            batch = []

            for file_or_path in enumerate_path(files):
                batch.append(file_or_path)
                if len(batch) == batch_size:
                    add_batch(batch)
                    batch = []

            if len(batch) > 0:
                add_batch(batch)

        if commit:
            repo.commit(message)


@data_commands.command('serve')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--host', '-h', default='127.0.0.1', required=False)
@click.option('--port', default=3030, required=False)
@click.pass_context
def serve_data(ctx, volumeid, host, port):
    from .server import app
    click.echo('Starting server on http://{host}:{port}'.format(host=host, port=port))

    with with_repo(ctx, volumeid) as repo:
        repo.close()
        app.config.update(dict(data_path=repo.data_path, repo_root=repo.repo_root, config_prefix=ctx.obj.config_prefix))
        app.run(host=host, port=port, debug=True)


def get_file_phase(split):
    rand = random.uniform(0, 1)

    phase = None
    for key, ranges in split.items():
        start_range, end_range = ranges

        if rand >= start_range < end_range:
            phase = key
            break

    return phase


def fill_in_vars(path, replace_vars):
    for var_name, var_value in replace_vars.items():
        path = path.replace('$' + var_name, str(var_value))

    return path


def safe_make_dirs(dir):
    try:
        os.makedirs(dir)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise


def path_elements(path):
    folders = []
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)

            break

    folders.reverse()

    return folders


def safe_rm_tree(path):
    try:
        shutil.rmtree(path)
    except OSError as ex:
        if ex.errno != errno.ENOENT:
            raise


@data_commands.command('clone')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--dest', '-d', required=True)
@click.option('--query', '-q', required=False)
@click.option('--delete', is_flag=True, required=False)
@click.pass_context
def clone_data(ctx, volumeid, dest, query, delete):
    dest = __expend_and_validate_path(dest, expand_vars=False, validate_path=False)

    def find_dest_root():
        elements = path_elements(dest)

        root = []
        for element in elements:
            if element.startswith('$'):
                break

            root.append(element)

        return os.path.join(*root)

    if delete:
        root_dest = find_dest_root()
        safe_rm_tree(root_dest)

    if '$phase' not in dest and '$@' not in dest:
        dest = os.path.join(dest, '$@')

    if '$name' not in dest and '$path' not in dest:
        dest = os.path.join(dest, '$name')

    split = {
        'train': (0.0, 0.7),
        'test': (0.7, 0.8),
        'validation': (0.8, 1.0),
    }

    with with_repo(ctx, volumeid, read_only=True) as repo:
        results = list(repo.metadata.query(query))
        with click.progressbar(length=len(results), show_pos=True) as bar:
            for item in results:
                _, data = repo.object_store.get_raw(item['id'])

                phase = get_file_phase(split)

                if phase is None:
                    continue

                item['phase'] = phase
                item['@'] = phase
                item['name'] = os.path.basename(item['path'])
                item['dir'] = os.path.dirname(item['path'])
                item['path'] = item['path']
                dest_file = fill_in_vars(dest, item)

                safe_make_dirs(os.path.dirname(dest_file))

                with open(dest_file, 'w') as f:
                    f.write(data)

                bar.update(1)


@data_commands.group('metadata')
def metadata_commands():
    pass


def stats_from_json(now, json):
    return os.stat_result((
        0,  # mode
        0,  # inode
        0,  # device
        0,  # hard links
        0,  # owner uid
        0,  # gid
        len(json),  # size
        0,  # atime
        now,
        now,
    ))


@metadata_commands.command('query')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.argument('query', nargs=1)
@click.pass_context
def query_metadata(ctx, query, path):
    with with_repo(ctx, path, read_only=True) as repo:
        for item in repo.query(query):
            print(os.path.join(repo.data_path, item['@sha']))


def chunks(l, n):
    result = []
    for item in l:
        result.append(item)

        if len(result) == n:
            yield result
            result = []

    if result:
        yield result


@metadata_commands.command('add')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--files', '-f', multiple=True)
@click.option('--data', '-d', required=False)
@click.option('--dataFile', '-df', required=False, type=click.File('rb'))
@click.option('--property', '-p', required=False, type=(str, str), multiple=True)
@click.option('--propertyInt', '-pi', required=False, type=(str, int), multiple=True)
@click.option('--propertyFloat', '-pf', required=False, type=(str, float), multiple=True)
@click.option('--update/--replace', is_flag=True, default=True, required=False)
@click.pass_context
def add_to_metadata(ctx, volumeid, files, data, datafile, property, propertyint, propertyfloat, update):
    if data is not None:
        if not isinstance(data, bytes):
            data = data.encode('utf8')

    now = time.time()

    if datafile:
        current_data = json.load(datafile)
    else:
        current_data = {}

        if data:
            data = data.decode('utf8')
            current_data = json.loads(data)

        for props in (property, propertyint, propertyfloat):
            if not props:
                continue

            for prop_name, prop_val in props:
                current_data[prop_name] = prop_val

    def filter_internal_data(full_current_data):
        return {key: val for key, val in full_current_data.items() if not key.startswith('@')}

    def compare_data(prev_data, key):
        prev_data_without_internal = filter_internal_data(prev_data) if prev_data else None

        item_current_data = current_data if not datafile else current_data[key]

        return prev_data_without_internal == item_current_data

    def update_data(prev_data, key):
        prev_data = prev_data or {}

        item_current_data = current_data if not datafile else current_data[key]

        prev_data.update(item_current_data)

        json_data_text = json.dumps(prev_data, sort_keys=True)

        data_blob = Blob()
        data_blob.data = json_data_text.encode('utf8')

        data_st = stats_from_json(now, json_data_text)

        return data_blob, prev_data, data_st

    def update_enumerate_function(current_batch):
        results = {sha: {} for sha in current_batch}
        for result in repo.metadata.get_head_data(current_batch):
            results[result['@sha']] = result

        for sha in current_batch:
            yield results[sha]

    def override_enumerate_function(current_batch):
        return [{}] * len(current_batch)

    enumerate_function = update_enumerate_function if update else override_enumerate_function

    def rel_path(file_enum):
        def rel_path_if_needed(path):
            if os.path.isabs(path):
                return os.path.relpath(path, repo.data_path)

            return path

        for current_file_path_chunks in file_enum:
            yield [rel_path_if_needed(current_file_path) for current_file_path in current_file_path_chunks]

    def enumerate_json():
        for item_key in current_data.keys():
            yield item_key

    with with_repo(ctx, volumeid) as repo:
        validate_repo_data_path(repo, volumeid)

        total_files = get_total_files_in_path(files) if not datafile else len(current_data)

        insert_batch_size = max(min(total_files // 20, 2000), 100)  # FIXME: hardcoded

        with click.progressbar(length=total_files, show_pos=True) as bar:
            index = repo.open_index()

            data_enumerator = enumerate_path(files) if not datafile else enumerate_json()

            for file_rel_path_chunk in rel_path(chunks(data_enumerator, insert_batch_size)):
                data_batch = []

                entries = {}

                for i, prev_json in enumerate(enumerate_function(file_rel_path_chunk)):
                    rel_file_path = file_rel_path_chunk[i]

                    if compare_data(prev_json, rel_file_path):
                        continue

                    blob, json_data, st = update_data(prev_json, rel_file_path)
                    data_batch.append((file_rel_path_chunk[i], json_data))

                    metadata_rel_path = rel_file_path + '.metadata'

                    metadata_tree_path = _fs_to_tree_path(metadata_rel_path)

                    entries[metadata_tree_path] = index_entry_from_stat(st, blob.id, 0)

                metadata = repo.metadata
                metadata.add_data(data=data_batch)
                index.set_entries(entries)

                bar.update(len(file_rel_path_chunk))



# coding=utf-8
try:
    from .backend_gcs_object_store import BackendGCSObjectStore
except ImportError:
    BackendGCSObjectStore = None

try:
    from .gcs_object_store import GCSObjectStore
except ImportError:
    GCSObjectStore = None



# coding=utf-8
from datetime import datetime


SERVER_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'  # ISO 8601 format
DISPLAY_DATETIME_FORMAT = '%b %d, %H:%M'


def format_datetime(server_datetime):
    d = datetime.strptime(server_datetime, SERVER_DATETIME_FORMAT)
    return datetime.strftime(d, DISPLAY_DATETIME_FORMAT)


SHORT_TEXT_MAX_LENGTH = 20
LONG_TEXT_MAX_LENGTH = 50


def create_text_truncator(max_length):
    def truncate_text(text):
        if max_length <= 0:
            return text

        text = ' '.join(text.splitlines())

        if len(text) > max_length:
            text = text[:max_length] + '...'

        return text

    return truncate_text


truncate_short_text = create_text_truncator(SHORT_TEXT_MAX_LENGTH)
truncate_long_text = create_text_truncator(LONG_TEXT_MAX_LENGTH)



# coding=utf-8
import click
import requests
import json

from mali_commands.commons import output_result
from mali_commands.options import project_id_option, experiment_id_option, metrics_option, model_weights_hash_option
from mali_commands.utilities.formatters import format_datetime, truncate_long_text, truncate_short_text


@click.group('experiments')
def experiments_commands():
    pass


@experiments_commands.command('list')
@project_id_option(required=True)
@click.pass_context
def list_experiments(ctx, projectid):
    """List experiments of a project.
    """
    list_experiments_path = 'projects/{project_id}/experiments'.format(project_id=projectid)
    result = ctx.obj.handle_api(ctx.obj, requests.get, list_experiments_path)
    experiments = result.get('experiments', [])

    displayed_fields = ['experiment_id', 'created_at', 'display_name', 'description']
    formatters = {
        'created_at': format_datetime,
        'display_name': truncate_short_text,
        'description': truncate_long_text,
    }
    output_result(ctx, experiments, displayed_fields, formatters=formatters)


@experiments_commands.command('updateMetrics')
@project_id_option(required=False)
@experiment_id_option(required=False)
@model_weights_hash_option(required=False)
@metrics_option(required=True)
@click.pass_context
def update_metrics(ctx, projectid, experimentid, weightshash, metrics):
    """Send experiment metrics to an experiment.

    There are 2 ways to identify the experiment: (1) specify both `--projectId` and `--experimentId`
    options or (2) specify `--weightsHash` option. When the model's weights hash is specified, mali
    would look up the experiment that the model belongs to.

    When both `--projectId` and `--experimentId` are specified, the `--weightsHash` option is ignored.

    Example:

    To send metrics to the 5th experiment of the project "123", run

    \b
        mali experiments update_metrics --projectId 123 --experimentId 5 --metrics '{"ex_cost": 99}'

    Or assuming that the model's weights hash "324e16b5e" was generated by this experiment at certain
    epoch or iteration, run

    \b
        mali experiments update_metrics --weightsHash 324e16b5e --metrics '{"ex_cost": 99}'
    """
    if projectid is not None and experimentid is not None:
        update_metrics_path = 'projects/{project_id}/experiments/{experiment_id}/metrics'\
            .format(project_id=projectid, experiment_id=experimentid)
    elif weightshash is not None:
        update_metrics_path = 'model_weights_hashes/{model_weights_hash}/metrics?experiment_only=1'\
            .format(model_weights_hash=weightshash)
    elif projectid is not None:
        raise click.BadOptionUsage('Please also specify --experimentId option.')
    elif experimentid is not None:
        raise click.BadOptionUsage('Please also specify --projectId option.')
    else:
        raise click.BadOptionUsage('Please specify the experiment using (1) --projectId and --experimentId options'
                                   'or (2) --weightsHash options.')

    data = _get_metrics_json(metrics)
    result = ctx.obj.handle_api(ctx.obj, requests.post, update_metrics_path, data)
    output_result(ctx, result, ['ok'])


@experiments_commands.command('updateModelMetrics')
@model_weights_hash_option(required=True)
@metrics_option(required=True)
@click.pass_context
def update_model_metrics(ctx, weightshash, metrics):
    """Send the model's metrics to its corresponding experiment epochs.

    Each model weights hash is attached to certain experiment epochs and thus can be used to send
    metrics that are relevant to those epochs. The model weights hash is a hexadecimal string. To
    calculate the weights hash of a model:

        - Calculate the sha1 strings of the weights for each layers

        - Calculate the sha1 string of the combine hashes. For example, the model has 3 layers with
    the layers' weight hashes: ['abc', '123', 'def'], the model weight hash is `sha1('abc123def')`

    Alternatively, you can use `missinglink-sdk` Python package to calculate the weights hash for
    your model. Each framework callback has its corresponding `get_weights_hash` method.

    WARNING: The same model weights hash can appear in different experiments or different epochs
    (for example, when running the same net twice). As such, this command will send the metrics to
    all these experiments/epochs that it can identify from the hash.

    Example:

    To send metrics of the model which hash is 324e16b5e, run

    \b
        mali experiments update_model_metrics --weightsHash 324e16b5e --metrics '{"ex_cost": 99}'

    """
    update_model_metrics_path = 'model_weights_hashes/{model_weights_hash}/metrics'\
        .format(model_weights_hash=weightshash)
    data = _get_metrics_json(metrics)

    result = ctx.obj.handle_api(ctx.obj, requests.post, update_model_metrics_path, data)
    output_result(ctx, result, ['ok'])


def _get_metrics_json(metrics_string):
    try:
        return json.loads(metrics_string)
    except ValueError:
        raise click.BadParameter('The metrics supplied is not a valid JSON dictionary format.')



# patch.py -- For dealing with packed-style patches.
# Copyright (C) 2009-2013 Jelmer Vernooij <jelmer@samba.org>
#
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as public by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Classes for dealing with git am-style patches.

These patches are basically unified diffs with some extra metadata tacked
on.
"""

from difflib import SequenceMatcher
import email.parser
import time

from .objects import (
    Blob,
    Commit,
    S_ISGITLINK,
    )

FIRST_FEW_BYTES = 8000


def write_commit_patch(f, commit, contents, progress, version=None,
                       encoding=None):
    """Write a individual file patch.

    :param commit: Commit object
    :param progress: Tuple with current patch number and total.
    :return: tuple with filename and contents
    """
    encoding = encoding or getattr(f, "encoding", "ascii")
    if isinstance(contents, str):
        contents = contents.encode(encoding)
    (num, total) = progress
    f.write(b"From " + commit.id + b" " +
            time.ctime(commit.commit_time).encode(encoding) + b"\n")
    f.write(b"From: " + commit.author + b"\n")
    f.write(b"Date: " +
            time.strftime("%a, %d %b %Y %H:%M:%S %Z").encode(encoding) + b"\n")
    f.write(("Subject: [PATCH %d/%d] " % (num, total)).encode(encoding) +
            commit.message + b"\n")
    f.write(b"\n")
    f.write(b"---\n")
    try:
        import subprocess
        p = subprocess.Popen(["diffstat"], stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE)
    except (ImportError, OSError):
        pass  # diffstat not available?
    else:
        (diffstat, _) = p.communicate(contents)
        f.write(diffstat)
        f.write(b"\n")
    f.write(contents)
    f.write(b"-- \n")
    if version is None:
        from dulwich import __version__ as dulwich_version
        f.write(b"Dulwich %d.%d.%d\n" % dulwich_version)
    else:
        f.write(version.encode(encoding) + b"\n")


def get_summary(commit):
    """Determine the summary line for use in a filename.

    :param commit: Commit
    :return: Summary string
    """
    return commit.message.splitlines()[0].replace(" ", "-")


def unified_diff(a, b, fromfile, tofile, n=3):
    """difflib.unified_diff that doesn't write any dates or trailing spaces.

    Based on the same function in Python2.6.5-rc2's difflib.py
    """
    started = False
    for group in SequenceMatcher(None, a, b).get_grouped_opcodes(n):
        if not started:
            yield b'--- ' + fromfile + b'\n'
            yield b'+++ ' + tofile + b'\n'
            started = True
        i1, i2, j1, j2 = group[0][1], group[-1][2], group[0][3], group[-1][4]
        sizes = "@@ -%d,%d +%d,%d @@\n" % (i1+1, i2-i1, j1+1, j2-j1)
        yield sizes.encode('ascii')
        for tag, i1, i2, j1, j2 in group:
            if tag == 'equal':
                for line in a[i1:i2]:
                    yield b' ' + line
                continue
            if tag == 'replace' or tag == 'delete':
                for line in a[i1:i2]:
                    if not line[-1:] == b'\n':
                        line += b'\n\\ No newline at end of file\n'
                    yield b'-' + line
            if tag == 'replace' or tag == 'insert':
                for line in b[j1:j2]:
                    if not line[-1:] == b'\n':
                        line += b'\n\\ No newline at end of file\n'
                    yield b'+' + line


def is_binary(content):
    """See if the first few bytes contain any null characters.

    :param content: Bytestring to check for binary content
    """
    return b'\0' in content[:FIRST_FEW_BYTES]


def shortid(hexsha):
    if hexsha is None:
        return b"0" * 7
    else:
        return hexsha[:7]


def patch_filename(p, root):
    if p is None:
        return b"/dev/null"
    else:
        return root + b"/" + p


def write_object_diff(f, store, old_file, new_file, diff_binary=False):
    """Write the diff for an object.

    :param f: File-like object to write to
    :param store: Store to retrieve objects from, if necessary
    :param old_file: (path, mode, hexsha) tuple
    :param new_file: (path, mode, hexsha) tuple
    :param diff_binary: Whether to diff files even if they
        are considered binary files by is_binary().

    :note: the tuple elements should be None for nonexistant files
    """
    (old_path, old_mode, old_id) = old_file
    (new_path, new_mode, new_id) = new_file
    old_path = patch_filename(old_path, b"a")
    new_path = patch_filename(new_path, b"b")

    def content(mode, hexsha):
        if hexsha is None:
            return Blob.from_string(b'')
        elif S_ISGITLINK(mode):
            return Blob.from_string(b"Submodule commit " + hexsha + b"\n")
        else:
            return store[hexsha]

    def lines(content):
        if not content:
            return []
        else:
            return content.splitlines()
    f.writelines(gen_diff_header(
        (old_path, new_path), (old_mode, new_mode), (old_id, new_id)))
    old_content = content(old_mode, old_id)
    new_content = content(new_mode, new_id)
    if not diff_binary and (
            is_binary(old_content.data) or is_binary(new_content.data)):
        f.write(b"Binary files " + old_path + b" and " + new_path +
                b" differ\n")
    else:
        f.writelines(unified_diff(lines(old_content), lines(new_content),
                     old_path, new_path))


# TODO(jelmer): Support writing unicode, rather than bytes.
def gen_diff_header(paths, modes, shas):
    """Write a blob diff header.

    :param paths: Tuple with old and new path
    :param modes: Tuple with old and new modes
    :param shas: Tuple with old and new shas
    """
    (old_path, new_path) = paths
    (old_mode, new_mode) = modes
    (old_sha, new_sha) = shas
    yield b"diff --git " + old_path + b" " + new_path + b"\n"
    if old_mode != new_mode:
        if new_mode is not None:
            if old_mode is not None:
                yield ("old mode %o\n" % old_mode).encode('ascii')
            yield ("new mode %o\n" % new_mode).encode('ascii')
        else:
            yield ("deleted mode %o\n" % old_mode).encode('ascii')
    yield b"index " + shortid(old_sha) + b".." + shortid(new_sha)
    if new_mode is not None:
        yield (" %o" % new_mode).encode('ascii')
    yield b"\n"


# TODO(jelmer): Support writing unicode, rather than bytes.
def write_blob_diff(f, old_file, new_file):
    """Write blob diff.

    :param f: File-like object to write to
    :param old_file: (path, mode, hexsha) tuple (None if nonexisting)
    :param new_file: (path, mode, hexsha) tuple (None if nonexisting)

    :note: The use of write_object_diff is recommended over this function.
    """
    (old_path, old_mode, old_blob) = old_file
    (new_path, new_mode, new_blob) = new_file
    old_path = patch_filename(old_path, b"a")
    new_path = patch_filename(new_path, b"b")

    def lines(blob):
        if blob is not None:
            return blob.splitlines()
        else:
            return []
    f.writelines(gen_diff_header(
        (old_path, new_path), (old_mode, new_mode),
        (getattr(old_blob, "id", None), getattr(new_blob, "id", None))))
    old_contents = lines(old_blob)
    new_contents = lines(new_blob)
    f.writelines(unified_diff(old_contents, new_contents,
                 old_path, new_path))


# TODO(jelmer): Support writing unicode, rather than bytes.
def write_tree_diff(f, store, old_tree, new_tree, diff_binary=False):
    """Write tree diff.

    :param f: File-like object to write to.
    :param old_tree: Old tree id
    :param new_tree: New tree id
    :param diff_binary: Whether to diff files even if they
        are considered binary files by is_binary().
    """
    changes = store.tree_changes(old_tree, new_tree)
    for (oldpath, newpath), (oldmode, newmode), (oldsha, newsha) in changes:
        write_object_diff(f, store, (oldpath, oldmode, oldsha),
                          (newpath, newmode, newsha), diff_binary=diff_binary)


def git_am_patch_split(f, encoding=None):
    """Parse a git-am-style patch and split it up into bits.

    :param f: File-like object to parse
    :param encoding: Encoding to use when creating Git objects
    :return: Tuple with commit object, diff contents and git version
    """
    encoding = encoding or getattr(f, "encoding", "ascii")
    contents = f.read()
    if (isinstance(contents, bytes) and
            getattr(email.parser, "BytesParser", None)):
        parser = email.parser.BytesParser()
        msg = parser.parsebytes(contents)
    else:
        parser = email.parser.Parser()
        msg = parser.parsestr(contents)
    return parse_patch_message(msg, encoding)


def parse_patch_message(msg, encoding=None):
    """Extract a Commit object and patch from an e-mail message.

    :param msg: An email message (email.message.Message)
    :param encoding: Encoding to use to encode Git commits
    :return: Tuple with commit object, diff contents and git version
    """
    c = Commit()
    c.author = msg["from"].encode(encoding)
    c.committer = msg["from"].encode(encoding)
    try:
        patch_tag_start = msg["subject"].index("[PATCH")
    except ValueError:
        subject = msg["subject"]
    else:
        close = msg["subject"].index("] ", patch_tag_start)
        subject = msg["subject"][close+2:]
    c.message = (subject.replace("\n", "") + "\n").encode(encoding)
    first = True

    body = msg.get_payload(decode=True)
    lines = body.splitlines(True)
    line_iter = iter(lines)

    for l in line_iter:
        if l == b"---\n":
            break
        if first:
            if l.startswith(b"From: "):
                c.author = l[len(b"From: "):].rstrip()
            else:
                c.message += b"\n" + l
            first = False
        else:
            c.message += l
    diff = b""
    for l in line_iter:
        if l == b"-- \n":
            break
        diff += l
    try:
        version = next(line_iter).rstrip(b"\n")
    except StopIteration:
        version = None
    return c, diff, version
